/*
 * Created on 16-Jun-2005
 * 
 */
package org.openrdf.jena;

import java.rmi.server.ServerCloneException;
import java.util.logging.Logger;

import org.openrdf.model.Resource;
import org.openrdf.model.Statement;
import org.openrdf.model.URI;
import org.openrdf.model.Value;
import org.openrdf.model.ValueFactory;
import org.openrdf.model.impl.StatementImpl;
import org.openrdf.repository.Repository;
import org.openrdf.repository.RepositoryResult;
import org.openrdf.sail.Sail;
import org.openrdf.sail.SailInitializationException;
import org.openrdf.repository.RepositoryException;
import org.openrdf.sail.StackableSail;
import org.openrdf.repository.config.RepositoryConfig;
import org.openrdf.repository.config.SailConfig;
import org.openrdf.repository.config.SailParameter;
import org.openrdf.repository.sail.SailRepository;

import info.aduna.collections.iterators.CloseableIterator;
import info.aduna.collections.iterators.EmptyIterator;
import info.aduna.collections.iterators.IteratorWrapper;
import info.aduna.iteration.CloseableIterationBase;

import com.hp.hpl.jena.graph.BulkUpdateHandler;
import com.hp.hpl.jena.graph.GraphUtil;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.graph.TripleMatch;
import com.hp.hpl.jena.graph.impl.AllCapabilities;
import com.hp.hpl.jena.graph.impl.GraphBase;
import com.hp.hpl.jena.shared.AddDeniedException;
import com.hp.hpl.jena.shared.DeleteDeniedException;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;

/**
 * This class wraps a Sesame repository and presents it as a Jena Graph
 * 
 * NOTE: This class does NOT support remove'ing triples from the iterators returned by find!
 * This means essentially everything, removing can only be done by performRemove call!
 * 
 * @author wf
 * @author grimnes
 * TODO: Think about and check if re-ification works.
 */
public class GraphSesame extends GraphBase {
	
	protected boolean writeable = true;
	
	protected Repository repository;
	//private org.openrdf.model.Graph sesameGraph;
	protected Logger logger = Logger.getLogger(GraphSesame.class.getName());
	
	private static ValueFactory sesameFactory;
	
	/**
	 * Get a Sesame ValueFactory used to create Sesame URI, blank node, literal, etc. 
	 * @return Sesame ValueFactory
	 */
	public static ValueFactory getValueFactory() { 
		return sesameFactory;
	}
	
	private void setCapabilities() { 
		capabilities=new AllCapabilities() {
			public boolean iteratorRemoveAllowed() { return false; }
		};
	}
	
	/**
	 * Create a Jena graph from an existing repository
	 *
	 */
	public GraphSesame(Repository repository) {		
		setCapabilities();
		this.repository=repository;
		try { 
			repository.initialize();
		} catch(Exception e) {
			//ignore
			//TODO: This is SHIT, why isn't there a repositoy.isInitialized?
		}
		
		sesameFactory=repository.getValueFactory();
	}
	
	/**
	 * Create a Jena Graph, which represents a Sesame in-memory local repository
	 * that does not have persistent support, synchroinzation support and inference capability.
	 */
	public GraphSesame() {
		this(false, false, null, null);
	}
	
	/**
	 * Create a Jena Graph, which represents a Sesame in-memory local repository 
	 * that has the following customizable features. It will firstly create a 
	 * Sesame repository, then create a Sesame graph from the repository.
	 * @param inferencing  the Sesame repository supports inference or not.   
	 * @param sync         the the Sesame repository is synchronized or not.
	 * @param fileName     the name of the file provides the persistent storage.
	 *                     Null means no persistent support required.
	 * @param format       the RDF data format in the persistent storage.
	 */
	public GraphSesame(boolean inferencing, boolean sync, 
			String fileName, String format) {
		super();
		setCapabilities();
		
		RepositoryConfig repConfig = new RepositoryConfig("Jena-Sesame");
		
		if (sync) {
			addSyncLayer(repConfig);
		}
		
		SailConfig memSail = null;
		if (inferencing)
			memSail = new SailConfig("org.openrdf.sesame.sailimpl.memory.MemoryStoreRDFSInferencer");
		else
			memSail = new SailConfig("org.openrdf.sesame.sailimpl.memory.MemoryStore");
		
		if (fileName != null) {
			if (!format.equals("rdfxml") && !format.equals("turtle"))
				format = "ntriples";
			
			memSail.addParameter("file", fileName);
			memSail.addParameter("dataFormat", format);
		}
		
		repConfig.addSailConfig(memSail);
		
		createOnConfig(repConfig);
	}
	
	
	/**
	 * Create a Jena Graph, which represents a Sesame native local repository.
	 * @param sync  the the Sesame repository is synchronized or not.
	 * @param dir   specifies the directory that can be used by the native sail to store its files.
	 */
	public GraphSesame(boolean sync, String dir) {
		super();
		setCapabilities();
		
		RepositoryConfig repConfig = new RepositoryConfig("Jena-Sesame");
		
		if (sync) {
			addSyncLayer(repConfig);
		}
		
		SailConfig nativeSail = 
			new SailConfig("org.openrdf.sesame.sailimpl.nativerdf.NativeStore");
		nativeSail.addParameter("dir", dir);
		
		repConfig.addSailConfig(nativeSail);
		
		createOnConfig(repConfig);
	}
	
	/**
	 * Create a Jena Graph, which represents a Sesame rdbms local repository.
	 * @param sync     true: the model is synchronized; false: not.
	 * @param driver   identifies the JDBC (Java Data Base Connectivity) driver that is to be used to access the database.
	 * @param url      identifies the location of the database through a URL.
	 * @param user     identifies a username with which Sesame can access the database. 
	 * @param password identifies a password with which Sesame can access the database.
	 */
	public GraphSesame(boolean sync, String driver, String url, String user, String password) {
		super();
		
		setCapabilities();
		
		RepositoryConfig repConfig = new RepositoryConfig("Jena-Sesame");
		
		if (sync) {
			addSyncLayer(repConfig);
		}
		
		SailConfig dbSail = new SailConfig("org.openrdf.sesame.sailimpl.rdbms.RdfRepository");
		dbSail.addParameter("jdbcDriver", driver);
		dbSail.addParameter("jdbcUrl", url);
		dbSail.addParameter("user", user);
		dbSail.addParameter("password", password);
		
		repConfig.addSailConfig(dbSail);
		
		createOnConfig(repConfig);
	}
	
	private void addSyncLayer(RepositoryConfig repConfig) {
		SailConfig syncSail = 
			new SailConfig("org.openrdf.sesame.sailimpl.sync.SyncRdfSchemaRepository");
		repConfig.addSailConfig(syncSail);
	}
	
	private void createOnConfig(RepositoryConfig repConfig) {
		repConfig.setWorldReadable(true);
		repConfig.setWorldWritable(true);     
		
		try {
			repository = _createRepository(repConfig);
			
			sesameFactory = repository.getValueFactory();
		} catch (Exception ex) {
			logger.severe("Can not create the Sesame repository");
			ex.printStackTrace();
			System.exit(1);
		}
	}
	
	
	/**
	 * Copied from sesame org.openrdf.sesame.server.Server
	 * @param rc
	 * @return a new repository configured as specified.
	 * @throws ServerConfigException
	 */
	private Repository _createRepository(RepositoryConfig rc) throws ServerConfigException {
		try {
			// Create and initialize the Sail stack
			Sail sailStack = null;
			
			for (SailConfig sailConfig : rc.getSailConfigStack()) {
				String className = sailConfig.getClassName();
				Class<?> sailClass = Class.forName(className);
				
				Sail sail = null;
				try {
					sail = (Sail)sailClass.newInstance();
				}
				catch (ClassCastException e) {
					throw new ServerConfigException(sailClass + " does not implement Sail interface.");
				}
				
				if (sailStack != null) {
					// Stack this Sail on top of the existing Sail stack
					try {
						((StackableSail)sail).setBaseSail(sailStack);
					}
					catch (ClassCastException e) {
						throw new ServerConfigException(className + " is not a StackableSail.");
					}
				}
				
				for (SailParameter param : sailConfig.getParameters()) {
					sail.setParameter(param.getName(), param.getValue());
				}
				
				sailStack = sail;
			}
			
			if (sailStack == null) {
				throw new ServerConfigException("No Sail stack has been specified");
			}
			
			Repository repository = new SailRepository(sailStack);
			repository.initialize();
			return repository;
		}
		catch (ClassNotFoundException e) {
			throw new ServerConfigException(e);
		}
		catch (InstantiationException e) {
			throw new ServerConfigException(e);
		}
		catch (IllegalAccessException e) {
			throw new ServerConfigException(e);
		}
		catch (RepositoryException e) {
			throw new ServerConfigException(e);
		}
	}
	
	// convert a jena triple to SesameTriple
	public Statement jena2sesame(Triple t) {
				
		Resource subject = null;
		URI predicate = null;
		Value object = null;
		try {
			subject = (Resource)ValueToNode.reverse(t.getSubject());		
			predicate = (URI)ValueToNode.reverse(t.getPredicate());
		} catch (ClassCastException e) {
			throw new IllegalArgumentException("You can't create triples with literals as subjects or predicates, or with BlankNodes as predicates.");
		}
		object = ValueToNode.reverse(t.getObject());
		
		return new StatementImpl(subject, predicate, object);
	}
	
	/**
	 * 
	 * 
	 * @see com.hp.hpl.jena.graph.impl.GraphBase#graphBaseFind(com.hp.hpl.jena.graph.TripleMatch)
	 */
	protected ExtendedIterator graphBaseFind(TripleMatch tm) {
		// throw new RuntimeException("GraphSesame::graphBaseFind");
		Statement stri=null;
		try {
			stri=jena2sesame(tm.asTriple());
		} catch (IllegalArgumentException ce) {
			// This might just mean that someone tried to make triples with blanknodes as predicates.
			// in a jena graph this is legal, but you wont find them here, 
			// return an empty iterator.
			return new ExtendedStatementIterator(new EmptyIterator<Statement>());
		}
		//logger.debug("\tJena: " + tm + "\n\tSesame: " + stri);
		try {
		RepositoryResult<Statement> stit = repository.getConnection().getStatements(
				stri.getSubject(), stri.getPredicate(), stri.getObject(),false, null);
		
		
		return new ExtendedStatementIterator(new IteratorWrapper<Statement>(stit.asCollection().iterator()));
		} catch (RepositoryException e) {
			throw new RuntimeException(e);
		}
	}
	
	
	/** 
	 * @see com.hp.hpl.jena.graph.impl.GraphWithPerform#performAdd(com.hp.hpl.jena.graph.Triple)
	 */
	public void performAdd(Triple t) {
		if (!writeable)
			throw new AddDeniedException("graph is readonly");
		Statement stri= jena2sesame(t);
//		logger.debug("\tJena: " + t + "\n\tSesame: " + stri);
		try {
			repository.getConnection().add(stri.getSubject(), stri.getPredicate(), stri.getObject());
		} catch (RepositoryException e) {
			throw new RuntimeException(e);
		}
	}
	
	/**
	 * @see com.hp.hpl.jena.graph.impl.GraphWithPerform#performDelete(com.hp.hpl.jena.graph.Triple)
	 */
	public void performDelete(Triple t) {
		if (!writeable)
			throw new DeleteDeniedException("graph is readonly");
		Statement stri= jena2sesame(t);
//		logger.debug("\tJena: " + t + "\n\tSesame: " + stri);
		try {
			repository.getConnection().remove(stri.getSubject(), stri.getPredicate(), stri.getObject());
		} catch (RepositoryException e) { 			
			throw new RuntimeException(e);
		}		   
	}
	
	/**  
	 * @see com.hp.hpl.jena.graph.Graph#queryHandler()
	 */
//	public QueryHandler queryHandler() {
//		if (queryHandler == null) 
//			queryHandler = new SesameQueryHandler(this);
//		return queryHandler;
//	}
//	
	public Repository getRepository() {
		return repository;
	}
	
	
	
	public void close() {
		try {
		logger.finer("closing a GraphSesame, also closes repository: "+repository);
		super.close();
		repository.shutDown();
		} catch (RepositoryException e){
			throw new RuntimeException(e);
		}
	}
	


	/**
	 * We override this, maybe Sesame's version is more efficient. 
	 *@Override 
	 */	
	protected int graphBaseSize() {
		//TODO: When the sesame guys fix this, use the native method.
		ExtendedIterator it = GraphUtil.findAll( this );
        int tripleCount = 0;
        while (it.hasNext()) { it.next(); tripleCount += 1; }
        it.close();
        return tripleCount; 
		
	}

	@Override
	public BulkUpdateHandler getBulkUpdateHandler() {
		return new SesameBulkUpdateHandler(this);
	}

	public boolean isWriteable() {
		return writeable;
	}

	public void setWriteable(boolean writeable) {
		this.writeable = writeable;
	}	
}