/*
 * Created on 25-Jun-2005
 * 
 */
package org.openrdf.jena;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.openrdf.model.BNode;
import org.openrdf.model.Literal;
import org.openrdf.model.Resource;
import org.openrdf.model.URI;
import org.openrdf.model.Value;
import org.openrdf.model.ValueFactory;


import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.rdf.model.AnonId;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.rdf.model.impl.LiteralImpl;
import com.hp.hpl.jena.rdf.model.impl.ModelCom;
import com.hp.hpl.jena.rdf.model.impl.PropertyImpl;
import com.hp.hpl.jena.rdf.model.impl.ResourceImpl;
import com.hp.hpl.jena.rdf.model.impl.StatementImpl;

/**
 * A Jena Node is equivalent to a Sesame Value. This class provides some
 * utilities to convert between Jena Nodes and Sesame Values.
 * @author wf
 * 
 */
public class ValueToNode {
    
    private static Logger logger = Logger.getRootLogger();
    
    private static ModelCom BogusModel = (ModelCom)ModelFactory.createDefaultModel();
    
    /**
     * Convert from Sesame Value to Jena Node.
     * Note: if null is passed in, null is passed back.
     * @param value a Sesame Value.
     * @return      a correspondent Jena Node.
     */
    static public Node convert(Value value) {
        Node n = null;
        if (value == null)
        	return null;
        if (value instanceof URI) {
            n = Node.createURI(value.toString());
        } else if (value instanceof BNode) {
            n = Node.createAnon(new AnonId(value.toString()));
        } else if (value instanceof Literal){
            n = Node.createLiteral(value.toString());
        } else {
            logger.fatal(value);
            String types;
            if (value != null)
            	types = value.getClass().getName();
            else
            	types = "null";
        	throw new RuntimeException("ValueToNode$convert: unknown value: "+value+" with unknown type:"
        			+types);
        }
        
        return n;
    }
    
    /**
     * Convert from Jena Node to Sesame Value
     * @param n a Jena Node.
     * @return  a correspondent Sesame Value.
     */
    static public Value reverse(Node n) {
    	Value value = null;
        ValueFactory factory = GraphSesame.getValueFactory();
        
        if (n.isBlank()) {
            // In Sesame Rio, the node can not begin with digits. But Jena blank node begins
            // with digits. We need to fix this.
            // A Sesame blank node looks like: node10jofjuktx143
            // A Jena blank node looks like: 1fe1feb:104f0e31293:-7ff9
            // We change it to: node1fe1febC104f0e31293CM7ff9
            // : becomes C, - becomes M
        	// prefix should be "node". Because when Sesame reads triples from the file, it will change
            // the name of any blank node to its own format.
            String s = n.getBlankNodeId().toString();
            String prefix = "node";
            if (s.startsWith(prefix)) 
            	value = factory.createBNode(s);
            else
                value = factory.createBNode(prefix + s.replaceAll(":", "C").replaceAll("\\-", "M"));
//            value = factory.createBNode(n.getBlankNodeId().toString());
        } else if (n.isURI()) {
            value = factory.createURI(n.getURI());
        } else if (n.isLiteral()) {
            value = factory.createLiteral(n.getLiteral().toString());
        } else if (n.matches(Node.ANY) || n.isVariable()) {
            value = null;
        } else {
        	logger.fatal(n);
            throw new RuntimeException("ValueToNode$reverse: unknown node type.");
        }
        
        return value;
    }
    
    /**
     * convert some sesame triples to Jena statement
     */
    public static Statement sesameToStatement(Resource subject, URI predicate, Value object)
    {
    	Node objNode = convert(object);
    	RDFNode objr = (objNode.isLiteral())? new LiteralImpl(objNode, BogusModel) : 
    		new ResourceImpl(objNode, BogusModel);
    	Statement s = new StatementImpl(
    			new ResourceImpl(convert(subject), BogusModel), 
    			new PropertyImpl(convert(predicate), BogusModel),
    			objr);
    	return s;
    }
    
    /**
     * convert some Sesame Statement to Jena statement
     */
    public static Statement sesameToStatement(org.openrdf.model.Statement statement)
    {
    	Node objNode = convert(statement.getObject());
    	RDFNode objr = (objNode.isLiteral())? new LiteralImpl(objNode, BogusModel) : 
    		new ResourceImpl(objNode, BogusModel);
    	Statement s = new StatementImpl(
    			new ResourceImpl(convert(statement.getSubject()), BogusModel), 
    			new PropertyImpl(convert(statement.getPredicate()), BogusModel),
    			objr);
    	return s;
    }
    
    /**
     * convert a jena statement to a sesame statement
     * @param statement
     * @return
     */
    public static org.openrdf.model.Statement jenaToSesameStatement(Statement statement)
    {
    	Value subject = reverse(statement.getSubject().asNode());
    	Value predicate = reverse(statement.getPredicate().asNode());
    	Value object = reverse(statement.getObject().asNode());
    	if (subject == null)
    		throw new RuntimeException("subject is null in statement "+statement);
    	if (predicate == null)
    		throw new RuntimeException("predicate is null in statement "+statement);
    	if (object == null)
    		throw new RuntimeException("object is null in statement "+statement);
    	if (!(subject instanceof Resource))
    		throw new RuntimeException("subject is not a resource in "+statement);
    	if (!(predicate instanceof URI))
    		throw new RuntimeException("predicate is not a URI in "+statement);
    		
    	org.openrdf.model.Statement result = new org.openrdf.model.impl.StatementImpl((Resource)subject,
    			(URI)predicate, object);
    	return result;
    }
    
    /**
     * transform a jena model to a collection of sesame statements,
     * this can be useful to mass-insert statements
     * @param input the input model
     * @return a collection of openrdf statements
     */
    public static Collection<org.openrdf.model.Statement> jenaModelToSesameStatements(Model input)
    {
    	// in the implementation, it is not clear how to initialise the size of a returning
    	// array or list, as the getSize() method's contract does not say that it is exactly the same
    	// size as the iterator, because of inference, etc.
    	ArrayList<org.openrdf.model.Statement> result = new ArrayList<org.openrdf.model.Statement>();
    	for (StmtIterator i = input.listStatements(); i.hasNext();)
    	{
    		org.openrdf.model.Statement s = jenaToSesameStatement(i.nextStatement());
    		result.add(s);
    	}
    	return result;
    }
    
//    /** Now this REALLY sucks **/
//	public static List<org.openrdf.model.Statement>castList(List<org.openrdf.model.Statement> in) {
//		List<org.openrdf.model.Statement> res=new Vector<org.openrdf.model.Statement>();
//		res.addAll(in);
//		return res;
//	}
}
